from PIL import Image
import time
import math
from numba import njit
import pygame
from pygame.locals import *

pygame.init()


"""
Génère le dégradé de couleurs.
Part de la couleur (r0, g0, b0)
puis passe par le blanc et va vers le noir.
seuil_fin permet de déterminer la proportion de couleurs
pour le dégradé blanc-noir.
"""
def faire_bande_couleurs(nmax=255, couleur_debut=(23, 50, 255), couleur_milieu=(255, 255, 255), couleur_fin=(0, 0, 0), seuil_fin=0.3):
    reste_fin = int(nmax*seuil_fin)
    r0, g0, b0 = couleur_debut
    r1, g1, b1 = couleur_milieu
    r2, g2, b2 = couleur_fin
    liste_couleurs = [(round(r0+(r1-r0)*(n/(nmax-reste_fin))**(0.5)),
                       round(g0+(g1-g0)*(n/(nmax-reste_fin))**(0.5)),
                       round(b0+(b1-b0)*(n/(nmax-reste_fin))**(0.5))) for n in range(1, nmax+2-reste_fin)]
    r, g, b = liste_couleurs[-1]
    for i in range(1, reste_fin+1):
        coeff = i/reste_fin
        liste_couleurs.append((round(r+(r2-r)*coeff), round(g+(g2-g)*coeff), round(b+(b2-b)*coeff)))
    return liste_couleurs

# On donne la liste des couleurs par lesquelles on veut passer.
# Les dégradés sont linéaires. On prendra le log de 'n' pour gérer le côté exponentiel du nombre d'itérations
def faire_couleurs(nmax=255, liste_departs=[(23, 50, 255), (23, 119, 0), (255, 84, 24), (15, 198, 255), (254, 245, 46), (171, 0, 168)]):
    liste = []
    liste_departs.append(liste_departs[0])
    for i in range(1, len(liste_departs)):
        #liste.extend(faire_bande_couleurs(round(nmax*(2**i)), liste_departs[i]))
        liste.extend(faire_bande_couleurs(nmax, liste_departs[i-1], (255, 255, 255), liste_departs[i]))
    return liste


# Pour voir la bande obtenue
def visualiser_couleurs(liste_couleurs):
    l = len(liste_couleurs)
    h = 150
    nouvelle_image = Image.new("RGB", (l, h))
    for i in range(l):
        for j in range(h):
            nouvelle_image.putpixel((i, j), liste_couleurs[i])
    nouvelle_image.show()

# Pour générer la liste des positions intermédiaires
# afin d'obtenir la trajectoire passant par les points clefs indiqués
def trajectoire(liste_positions):
    """ Les positions sont (x, y, zoom) """
    traj = []
    for i in range(1, len(liste_positions)):
        p1, z1 = liste_positions[i-1]
        p2, z2 = liste_positions[i]
        x1, y1 = p1
        x2, y2 = p2
        d = z2 - z1
        def f(n):
            return (0.95**n-0.95**d)/(1-0.95**d)
        for j in range(d):
            x = x2 + (x1-x2)*f(j)
            y = y2 + (y1-y2)*f(j)
            traj.append(((x, y), z1+j))
    traj.append(liste_positions[-1])
    return traj

# On utilise un cache pour accélérer les itérations
@njit
def convergence(c, nmax):
    n = 0
    z = 0
    while n < nmax and abs(z) < 2:
        z = z*z + c
        n += 1
    return n

# On prend le log de n pour transformer en linéaire une croissance exponentielle
def determine_couleur(n, liste_couleur):
    return liste_couleurs[round(math.log(1+n/255, 2)*255)%len(liste_couleurs)]

# Permet de trouver le bon zoom à partir de la taille trouvée dans Xaos
def calcule_zoom(cible):
    return math.log(cible/2)/math.log(0.95)

# La fonction qui calcule les images
# nmax : itération maximale
# centre : centre de l'image
# l : largeur de l'image, en pixels
# h : hauteur de l'image en pixels
# zoom : facteur de zoom
# k : facteur d'overscaling (pour calculer à une plus grande résolution)
# liste_couleurs : pour passer une liste de couleurs différente
# chunk : pour calculer par gros bloccs
# premier : permière passe ou pas
# image : l'image déjà calculée
def mandelbrot(nmax=255, centre = (-.5, 0), l=500, h=500, zoom=0, k=1,
               nom="mandelbrot.png", enregistrer=False, liste_couleurs=None, chunk=1, premier=True, image = None):
    print('nmax :', nmax)
    xc, yc = centre
    R = nmax**2
    l = l*k
    h = h*k
    echelle=(1/(0.5*l))*0.95**zoom
    xmin = xc - 0.5*l*echelle
    xmax = xc + 0.5*l*echelle
    ymin = yc - 0.5*h*echelle
    ymax = yc + 0.5*h*echelle
    dx = (xmax-xmin)/l
    dy = (ymax-ymin)/h
    #print(zoom, echelle, xmin, xmax, xmax-xmin)
    if image == None:
        nouvelle_image = Image.new("RGB", (l, h))
    else:
        nouvelle_image = image
    if liste_couleurs is None:
        liste_couleurs = faire_couleurs(255)
        #visualiser_couleurs(liste_couleurs)
    n_min = nmax
    n_max = 0
    matrice = [[0 for _ in range(l)] for _ in range(h)]
    limites = [255]
    while limites[-1] < nmax:
        limites.append(limites[-1]*2)
    repartition = [0] * (len(limites)+1)
    nb_convergents = 0
    nb_au_dessus_seuil = 0
    for i in range(0, l, chunk):
        for j in range(0, h, chunk):
            if premier or i%(2*chunk)!=0 or j%(2*chunk)!=0:
                x = xmin+i*dx
                y = ymax-j*dy
                c = complex(x, y)
                n = convergence(c, nmax)
                matrice[i][j] = n
                interv = 0
                while interv < len(limites) and limites[interv] <= n:
                    interv += 1
                repartition[interv] += 1
                if n < nmax:
                    n_min = min(n_min, n)
                    n_max = max(n_max, n)
                    coul = determine_couleur(n, liste_couleurs)
                    nb_convergents += 1
                    nb_au_dessus_seuil += (2*n > nmax)
                else:
                    coul = (0, 0, 0)
                for di in range(chunk):
                    for dj in range(chunk):
                        nouvelle_image.putpixel((i+di, j+dj), coul)    
    #if k > 1:
    #    nouvelle_image = nouvelle_image.resize((l//k, h//k), Image.Resampling.LANCZOS)
    # Pour regarder le nombre d'itérations nécessaires au min et au max
    print((n_min, n_max))
    #nouvelle_image.show()
    return nouvelle_image, nb_au_dessus_seuil / max(1, nb_convergents), echelle

#visualiser_couleurs(faire_couleurs(nmax=1000))

"""
La partie animation est à revoir
pour que les vitesses de déplacement
soient moins importantes au début.
Et éventuellement pour permettre de placer plusieurs points de passages.
"""
def faire_animation(traj, t, k, nmax, nom='mandelbrot.gif', debut=0, liste_couleurs=None, adapter_nmax=True):
    #x1, y1 = centre1
    #x2, y2 = centre2
    liste_images = []
    nmax0 = 255
    #liste_couleurs = faire_couleurs(nmax=255)
    #print(len(liste_couleurs))
    #visualiser_couleurs(liste_couleurs)
    for pt, zoom in traj[debut:]:
        debut = time.time()
        image, nb_au_dessus_seuil, echelle = mandelbrot(nmax=nmax, centre=pt, l=t, h=t, zoom=zoom, k=k,
                                               liste_couleurs=liste_couleurs, nom=f'tmp/mandelbrot-{zoom:03}.png',enregistrer=True)
        if adapter_nmax:
            print("au dessus de seuil :", nb_au_dessus_seuil)
            if nb_au_dessus_seuil > 0.0015:
                nmax *= 2
                print("nouvelle valeur de nmax :", nmax)
            elif nb_au_dessus_seuil < 0.0002 and nmax > nmax0:
                nmax //= 2
                print("nouvelle valeur de nmax :", nmax)
        print(zoom, time.time()-debut)
        liste_images.append(image)
    #liste_images[0].save(nom,
    #           save_all=True, append_images=liste_images[1:], optimize=False, duration=60, loop=0)
    
nmax = 255*(2**2)  # nombre d'itérations maximum avant de considérer que le point est dans l'ensemble.
# Plus nmax est grand, plus les frontières sont précises, mais plus c'est lent.
k = 1  # Calcule une image k fois plus grande pour lisser. Multiplie le temps de calcul par k^2.
t = 1024  # taille en pixels. Sert pour la longueur et la largeur de l'image.

w, h = t, t
screen = pygame.display.set_mode((w, h)) 
centre, zoom = (-0.51, 0), 0  # image de base

liste_couleurs = faire_couleurs(nmax=255) #, liste_departs=[(150,150, 255)])
#visualiser_couleurs(liste_couleurs[:1000])
animer = False # Pour décider si on fait une seule image ou une animation

debut = time.time()  # Pour mesurer le temps.
if not animer:
    centre, zoom = (-0.51, 0), 0  # image de base
    #centre, zoom = (-0.754,0.05), 150 # il y a des trucs à creuser là
    #centre, zoom = (-0.7540001879,0.05005001206), 600
    #centre, zoom = (-0.753512,0.050045), 225  # c'est bien
    #centre, zoom = (-0.753515,0.0500415), 190 # c'est bien
    #centre, zoom = (-0.754323, 0.050343), 210  #185 c'est nickel
    #centre, zoom = (-1.7863453469975, 0), 479
    #centre, zoom = (-1.2568227, -.3797), 250 # des trucs à creuser ici
    #centre, zoom = (-1.256826769, -.379700496), 328 # magnifique
    centre, zoom = (-1.2568267683133112, -0.37970049605555086), 469 # doubles spirales
    #centre, zoom = (-1.25682676847, -0.37970049621), 428 # vaisseau spacial
    #centre, zoom = (-1.2568267682710, -0.3797004961061772), 512 # jolie symétrie
    #centre, zoom = (-1.2568267682707852, -0.37970049610604956), 605 # joli
    #centre, zoom = (-1.2576484249526123, 0.37806699089099255), 200
    #centre, zoom = (-1.4796619997106286, 0.0010322310047319511), 550 # il y a des trucs
    #centre, zoom = (-0.4592700478034031, -0.582131910623331), 520
    #centre, zoom = (-1.2568268213976808, -0.3797004927037099), 410
    #centre, zoom = (-0.174660499607, 1.071260050166), 320
    #centre, zoom = (-0.691615716122, -0.346481919731), calcule_zoom(0.005836284875)
    running = True
    restart = True
    facteur_zoom = 10
    clics = [False, False, False] # gauche, milieu, droit
    
    surface_souris = pygame.Surface((w, h)).convert_alpha()
    surface_image = pygame.Surface((w*k, h*k))
    compteur = 1
    
    while running:
        x, y = pygame.mouse.get_pos()
        surface_souris.fill((0, 0, 0, 0))   
        
        for event in pygame.event.get(): 
        
            if event.type == QUIT:  # Pour quitter
                running = False
            elif event.type == pygame.KEYDOWN:  # Touches du clavier
                if event.key == pygame.K_KP_PLUS: # + -> zoom avant
                    restart = True
                    zoom += facteur_zoom
                elif event.key == pygame.K_KP_MINUS: # - -> zoom arrière
                    restart = True
                    zoom -= facteur_zoom
                elif event.key == pygame.K_UP: # on double la limite
                    restart = True
                    nmax *= 2
                elif event.key == pygame.K_DOWN: # on divise par 2 la limite
                    restart = True
                    nmax //= 2
                elif event.key == pygame.K_r:  # on revient au début
                    restart = True
                    centre, zoom = (-0.51, 0), 0
                elif event.key == pygame.K_s:  # on enregistre l'image
                    if k > 1:
                        nouvelle_image = img.resize((w, h), Image.Resampling.LANCZOS)
                    else:
                        nouvelle_image = img
                    nouvelle_image.save(f"mandelbrot_{compteur}.png")
                    compteur += 1
            elif event.type == pygame.MOUSEBUTTONDOWN:  # clic souris
                if event.button == 1:  # clic gauche
                    clics[0] = True
                    x1, y1 = x, y
                elif event.button == 2:  # clic milieu
                    clics[1] = True
                elif event.button == 3: # clic droit
                    clics[2] = True
            elif event.type == pygame.MOUSEBUTTONUP:  # on lâche le clic
                if event.button == 1:  # clic gauche
                    clics[0] = False
                    taille = min(abs(x2-x1), abs(y2-y1))
                    if taille > 2:
                        #print(f"delta zoom {calcule_zoom(min(abs(x2-x1), abs(y2-y1)))}")
                        zoom += calcule_zoom(2*min(abs(x2-x1)/w, abs(y2-y1)/h))
                        xc, yc = centre
                        #print("points", (x1, y1), (x2, y2))
                        #print("brut", (x1+x2)/(2), (y1+y2)/(2))
                        #print("echelle", echelle*(x1+x2)/(2), echelle*(y1+y2)/(2))
                        centre = xc+k*echelle*(x1+x2-w)/2, yc-k*echelle*(y1+y2-h)/(2)
                        #print("Nouveau centre", centre)
                        restart = True
                elif event.button == 2:  # clic milieu
                    clics[1] = False
                elif event.button == 3: # clic droit
                    clics[2] = False
                print(event.button)
        
        if clics[0]:
            x2, y2 = x, y
            
            pygame.draw.rect(surface_souris, (200, 200, 200, 200), pygame.Rect(min(x1, x2), min(y1, y2), abs(x2-x1), abs(y2-y1)))
            
                    
        if restart:
            restart = False
            chunk = 16
            premier = True
            img = Image.new("RGB", (w*k, h*k))
            print(f"centre, zoom = {centre}, {zoom}")
                    
        if chunk > 0:
            print("on calcule avec chunk = ", chunk)
            debut = time.time()
            img, nb, echelle = mandelbrot(nmax=nmax, centre=centre, l=t, h=t, zoom=zoom, k=k,
                                 liste_couleurs=liste_couleurs, chunk=chunk, premier=premier, image=img)
            print("temps", time.time()-debut)
            #print(nb)
            premier=False
            chunk = chunk // 2
            #img.show()
            py_image = pygame.image.fromstring(img.tobytes(), img.size, img.mode)
            if chunk == 0:
                print("fini")
        surface_image.blit(py_image, (0, 0))
        screen.blit(pygame.transform.scale(surface_image, (w, h)), (0, 0))
        screen.blit(surface_souris, (0, 0))
        pygame.display.flip()
    pygame.quit()

else:  # Les animations
    # Définition du centre et du niveau de zoom
    # le niveau de zoom peut aller de -10 à environ 600
    # Vous trouverez ci-dessous plusieurs valeurs "intéressantes"
    centre, zoom = (-0.51, 0), 0  # image de base
    points = []
    points.append((centre, zoom))
    centre, zoom = (-1.764, 0), 78  # auto-similarité
    points.append((centre, zoom))
    centre, zoom = (-1.78603, 0), 155  # autre auto-similarité
    points.append((centre, zoom))
    centre, zoom = (-1.7863438, 0), 260  # autre auto-similarité
    points.append((centre, zoom))
    #centre, zoom = (-1.786343, 0), 230  # Il faut augmenter le nmax à 1000
    #points.append((centre, zoom))
    centre, zoom = (-1.7863453385, 0), 365  # auto-similarité
    points.append((centre, zoom))
    centre, zoom = (-1.78634534685, 0), 444  # auto-similarité
    points.append((centre, zoom))
    centre, zoom = (-1.7863453469975, 0), 523  # auto-similarité
    points.append((centre, zoom))

    #points = [((-0.51, 0), 0), ((-0.509999960095569,0.5150000450032913), 630)] # pour le zoom en haut
    #points = [((-0.51, 0), 0), ((-0.50941423032647,0.5153510027998385), 620)] #zoom dans spirale
    #points = [((-0.51, 0), 0), ((-0.551815098957,0.62568476913), 460)] #zoom vers spirale
    points = [((-0.51, 0), 0), ((-0.7540001879,0.0500001206), 600)] # dans les zones entre les 2 premiers disques
    #centre, zoom = (-0.509999960095569,0.5150000450032913), 500  # Vous pouvez zoomer fort ici
    # zoom avec les très jolis passage
    points = [((-0.51, 0), 0),
              ((-1.256826769, -.379700496), 328),
              ((-1.2568267682710, -0.3797004961061772), 512),
              ((-1.2568267682707852, -0.37970049610604956), 605)]
    # zoom vers les zones "lumineuses"
    points = [((-0.51, 0), 0),
              ((-1.2576484249526123, 0.37806699089099255), 200),
              ((-1.2576484249526123, 0.37806699089099255), 540)]

    # mouais
    points = [((-0.51, 0), 0),
              ((-1.4796619997106286, 0.0010322310047319511), 550)]

    points = [((-0.51, 0), 0),
              ((-0.459270047792,-0.582131910626), 600)]

    #print("calcul de la trajectoire")
    traj = trajectoire(points)
    #nmax = 130560
    #print("fin du calcul de la trajectoire")
    centre_init, zoom_init = (-.51, 0), 0  # le centre et le zoom de départ
    # On va zoomer jusqu'à arriver au centre final
    faire_animation(traj, t, k, nmax,
                    nom='mandelbrot.gif', liste_couleurs=liste_couleurs,
                    adapter_nmax=False)